// © 2023-2024 John Breaux
//See LICENSE.md for license
//! Assembles a binary using the given [AST](crate::parser::ast)

use error::{AResult, ErrorKind::*};
use std::collections::HashMap;

use crate::{assembler::canonical::Canonicalize, lexer::token, parser::ast::*, span::Span};

use self::error::{Error, ErrorKind};

/// Assembles a binary using the given [Assemble]-able item
#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct Assembler<'t> {
    /// The assembled output
    output: Vec<u16>,
    /// Table of labels, for backpatching
    labels: HashMap<&'t str, usize>,
    /// Backpatch table for jump instructions
    jump_queue: Vec<(usize, &'t str)>,
    /// Backpatch table for immediate values
    expr_queue: Vec<(usize, Expr<'t>)>,
    /// Base address from .org directives
    org_base: usize,
    /// Last seen index in input
    loc: Span<usize>,
}

impl<'t> Assembler<'t> {
    pub fn new() -> Self {
        Default::default()
    }
    pub fn assemble<T: Assemble<'t>>(&mut self, t: &T) -> AResult<&mut Self> {
        t.assemble_in(self)
    }
    /// Gets the address of a label
    pub fn addrof(&self, name: &str) -> Option<u16> {
        self.labels.get(name).map(|v| *v as u16)
    }
    /// Gets the value at a label
    pub fn valueof(&self, name: &str) -> Option<u16> {
        self.output.get(self.addrof(name)? as usize).copied()
    }
    fn push(&mut self, word: u16) {
        self.output.push(word)
    }
    fn error(&self, kind: ErrorKind) -> Error {
        Error { span: self.loc, kind }
    }
    /// Backpatches everything, and yoinks the output buffer.
    pub fn out(&mut self) -> AResult<Vec<u16>> {
        // Resolve jumps
        for (idx, key) in &self.jump_queue {
            // eprintln!("Patching jump at {idx} with key {key}");
            match self.labels.get(key).map(|addr| addr.wrapping_sub(*idx as _) as i16) {
                None => Err(self.error(UndefinedLabel(key.to_string())))?,
                Some(value @ -0x3ff..=0x3fc) => self.output[*idx] |= (value - 1) as u16 & 0x3ff,
                Some(value) => Err(self.error(LongJump(value)))?,
            }
        }
        // Resolve immediates through late expression evaluation.
        for (idx, expr) in &self.expr_queue {
            // eprintln!("Patching immediate at {idx} with expression {expr:?}");
            self.output[*idx] = self.eval(expr)?;
        }
        let out = std::mem::take(&mut self.output);
        *self = Default::default();
        Ok(out)
    }

    pub fn add_label(&mut self, label: &'t str) -> AResult<()> {
        if *self.labels.entry(label).or_insert(self.output.len()) != self.output.len() {
            Err(self.error(RedefinedLabel(label.into())))?
        }
        Ok(())
    }

    /// Appends an expr as an extword, deferring its calculation for later
    pub fn defer_expr(&mut self, e: Expr<'t>) {
        self.expr_queue.push((self.output.len(), e));
        self.push(0);
    }
    /// Defers resolution of a jump label until output time
    /// The jump label will be later resolved to the NEXT word.
    pub fn defer_jump(&mut self, label: &'t str) {
        self.jump_queue.push((self.output.len(), label))
    }
}

pub trait Assemble<'t> {
    fn assemble(&self) -> AResult<Vec<u16>> {
        self.assemble_in(&mut Default::default())?.out()
    }
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>>;
}

impl<'t> Assemble<'t> for Statements<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        for stmt in &self.stmts {
            stmt.assemble_in(a)?;
        }
        Ok(a)
    }
}
impl<'t> Assemble<'t> for Statement<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        match self {
            Statement::Label(label) => a.add_label(label).map(|_| a),
            Statement::Insn(i) => i.assemble_in(a),
            Statement::Directive(d) => d.assemble_in(a),
            Statement::Comment(_) => Ok(a),
        }
    }
}
impl<'t> Assemble<'t> for Directive<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        match self {
            Directive::Define(_) => {}
            Directive::Org(base) => a.org_base = a.eval(base)? as usize,
            Directive::Word(expr) => a.defer_expr(*expr.clone()),
            Directive::Words(exprs) => {
                for expr in exprs {
                    a.defer_expr(expr.clone())
                }
            }
            Directive::String(str) => {
                str.assemble_in(a)?;
            }
        }
        Ok(a)
    }
}

impl<'t> Assemble<'t> for &'t str {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        for chunk in self.as_bytes().chunks(2) {
            match chunk.len() {
                0 => a.push(0),
                1 => {
                    a.push(chunk[0] as u16);
                    return Ok(a);
                }
                2 => a.push((chunk[1] as u16) << 8 | chunk[0] as u16),
                n => unreachable!("expected chunks of length 2, got length {n}"),
            }
        }
        a.push(0);
        Ok(a)
    }
}

impl<'t> Assemble<'t> for Instruction<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        let Self { span, kind } = self;
        a.loc = *span;
        kind.assemble_in(a)
    }
}
impl<'t> Assemble<'t> for InstructionKind<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        match self {
            InstructionKind::NoEm(v) => v.assemble_in(a),
            InstructionKind::OneEm(v) => v.assemble_in(a),
            InstructionKind::OneArg(v) => v.assemble_in(a),
            InstructionKind::TwoArg(v) => v.assemble_in(a),
            InstructionKind::Jump(v) => v.assemble_in(a),
            InstructionKind::Reti(v) => v.assemble_in(a),
            InstructionKind::Br(v) => v.assemble_in(a),
        }
    }
}
impl<'t> Assemble<'t> for NoEm {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        eprintln!(
            "Warning: directly assembling a noncanonical instruction may lead to unwanted overhead"
        );
        self.clone().to_canonical().assemble_in(a)
    }
}
impl<'t> Assemble<'t> for OneEm<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        eprintln!(
            "Warning: directly assembling a noncanonical instruction may lead to unwanted overhead"
        );
        self.clone().to_canonical().assemble_in(a)
    }
}
impl<'t> Assemble<'t> for OneArg<'t> {
    /// `[ 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0 ]`
    /// `[  0  0  0  1  0  0 [op:3  ] bw [Ad ] [dst_reg:4] ]`
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        let Self { opcode, width, src } = self;
        let (src_reg, src_mode, src_ext) = source(src);
        a.push(
            0b000100 << 10 | one_arg(*opcode) << 7 | (*width as u16) << 6 | src_mode << 4 | src_reg,
        );
        if let Some(expr) = src_ext {
            a.defer_expr(expr)
        }
        Ok(a)
    }
}
impl<'t> Assemble<'t> for TwoArg<'t> {
    /// `[ 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0 ]`
    /// `[ [opcode:4 ] [src_reg:4] Ad bw [As ] [dst_reg:4] ]`
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        let Self { opcode, width, src, dst } = self;
        let (src_reg, src_mode, src_ext) = source(src);
        let (dst_reg, dst_mode, dst_ext) = destination(dst);
        a.push(
            two_arg(*opcode) << 12
                | src_reg << 8
                | dst_mode << 7
                | (*width as u16) << 6
                | src_mode << 4
                | dst_reg,
        );

        if let Some(expr) = src_ext {
            a.defer_expr(expr)
        }
        if let Some(expr) = dst_ext {
            a.defer_expr(expr)
        }
        Ok(a)
    }
}
impl<'t> Assemble<'t> for Jump<'t> {
    /// `[ 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0 ]`
    /// `[  0  0  1 [cond:3] +- [word_offset:10          ] ]`
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        let Self { opcode, dst } = self;
        let word = 1 << 13
            | jump(*opcode) << 10
            | match *dst {
                JumpDst::Rel(value) if value & 1 == 1 => return Err(a.error(OddJump(value))),
                JumpDst::Rel(value) if !(-0x3fe..=0x400).contains(&value) => {
                    return Err(a.error(LongJump(value)))
                }
                JumpDst::Rel(value) => (value - 1) as u16 >> 1 & 0x3ff,
                JumpDst::Label(label) => {
                    a.defer_jump(label);
                    0
                }
            } & 0x3ff;
        a.push(word);
        Ok(a)
    }
}
impl<'t> Assemble<'t> for Reti {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        a.output.push(0b0001_0011_0000_0000);
        Ok(a)
    }
}
impl<'t> Assemble<'t> for Br<'t> {
    fn assemble_in<'a>(&self, a: &'a mut Assembler<'t>) -> AResult<&'a mut Assembler<'t>> {
        eprintln!(
            "Warning: directly assembling a noncanonical instruction may lead to unwanted overhead"
        );
        self.clone().to_canonical().assemble_in(a)
    }
}

pub fn one_arg(opcode: token::OneArg) -> u16 {
    opcode as u16
}

pub fn two_arg(opcode: token::TwoArg) -> u16 {
    opcode as u16 + 4
}

pub fn jump(opcode: token::Jump) -> u16 {
    use token::Jump;
    match opcode {
        Jump::Jne | Jump::Jnz => 0,
        Jump::Jeq | Jump::Jz => 1,
        Jump::Jnc | Jump::Jlo => 2,
        Jump::Jc | Jump::Jhs => 3,
        Jump::Jn => 4,
        Jump::Jge => 5,
        Jump::Jl => 6,
        Jump::Jmp => 7,
    }
}

/// Returns a tuple of (Reg, AddrMode, extword)
pub fn source<'t>(src: &Src<'t>) -> (u16, u16, Option<Expr<'t>>) {
    use SrcSpecial::*;
    match src {
        Src::Special(Four) => (2, 2, None),
        Src::Special(Eight) => (2, 3, None),
        Src::Special(Zero) => (3, 0, None),
        Src::Special(One) => (3, 1, None),
        Src::Special(Two) => (3, 2, None),
        Src::Special(NegOne) => (3, 3, None),
        Src::Immediate(e) => (0, 3, Some(*e.clone())),
        Src::Absolute(e) => (2, 1, Some(*e.clone())),
        Src::Direct(r) => (*r as u16, 0, None),
        Src::Indexed(e, r) => (*r as u16, 1, Some(*e.clone())),
        Src::Indirect(r) => (*r as u16, 2, None),
        Src::PostInc(r) => (*r as u16, 3, None),
        Src::BareExpr(e) => (0, 3, Some(*e.clone())),
    }
}
/// Returns a tuple of (Reg, AddrMode, Extword)
pub fn destination<'t>(dst: &Dst<'t>) -> (u16, u16, Option<Expr<'t>>) {
    use DstSpecial::*;
    match dst {
        Dst::Special(Zero) => (3, 0, None),
        Dst::Special(One) => (3, 1, None),
        Dst::Absolute(e) => (2, 1, Some(*e.clone())),
        Dst::Indexed(e, r) => (*r as u16, 1, Some(*e.clone())),
        Dst::Direct(r) => (*r as u16, 0, None),
    }
}

impl<'t> Assembler<'t> {
    /// Evaluates an [Expr] using labels and constants defined in the current program
    fn eval(&self, expr: &Expr) -> AResult<u16> {
        match expr {
            Expr::Binary(head, tails) => {
                let mut head = self.eval(head)?;
                for (op, tail) in tails {
                    let tail = self.eval(tail)?;
                    head = match op {
                        BinOp::Mul => head.wrapping_mul(tail),
                        BinOp::Div => head.wrapping_div(tail),
                        BinOp::Rem => head.wrapping_rem(tail),
                        BinOp::Add => head.wrapping_add(tail),
                        BinOp::Sub => head.wrapping_sub(tail),
                        BinOp::Lsh => head.wrapping_shl(tail as u32),
                        BinOp::Rsh => head.wrapping_shr(tail as u32),
                        BinOp::And => head & tail,
                        BinOp::Xor => head ^ tail,
                        BinOp::Or => head | tail,
                    };
                }
                Ok(head)
            }
            Expr::Unary(ops, tail) => {
                let mut tail = self.eval(tail)?;
                for op in ops {
                    tail = match op {
                        UnOp::Not => !tail,
                        UnOp::Neg => 0u16.wrapping_sub(tail),
                        UnOp::Deref => *self
                            .output
                            .get(tail.wrapping_sub(self.org_base as u16) as usize >> 1)
                            .ok_or_else(|| self.error(OobRead(tail)))?,
                    }
                }
                Ok(tail)
            }
            Expr::Group(e) => self.eval(e),
            Expr::Number(n) => Ok(*n),
            Expr::Ident(name) => {
                self.valueof(name).ok_or_else(|| self.error(UndefinedLabel(name.to_string())))
            }
            Expr::AddrOf(name) => self
                .addrof(name)
                .map(|p| (p << 1).wrapping_add(self.org_base as u16))
                .ok_or_else(|| self.error(UndefinedLabel(name.to_string()))),
        }
    }
}

pub mod error {
    use std::fmt::Display;

    use crate::span::Span;

    pub type AResult<T> = Result<T, Error>;

    #[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct Error {
        pub span: Span<usize>,
        pub kind: ErrorKind,
    }
    impl std::error::Error for Error {}

    #[derive(Clone, Debug, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub enum ErrorKind {
        #[default]
        Todo,
        /// A label was used, but not defined
        UndefinedLabel(String),
        RedefinedLabel(String),
        OobRead(u16),
        OddJump(i16),
        LongJump(i16),
        /// A plethora of [Error]s
        Errors(Vec<Error>),
    }
    impl Display for Error {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            let Self { kind, span } = self;
            write!(f, "[{span}]: ")?;
            write!(f, "Error: {kind}")
        }
    }
    impl Display for ErrorKind {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                ErrorKind::Todo => write!(f, "Not yet implemented"),
                ErrorKind::UndefinedLabel(label) => write!(f, "Label '{label}' not defined"),
                ErrorKind::RedefinedLabel(label) => write!(f, "Label '{label}' already defined"),
                ErrorKind::OobRead(addr) => {
                    write!(f, "Out of bounds read in constant expression: {addr}")
                }
                ErrorKind::OddJump(to) => write!(f, "Cannot jump to odd location: {to}"),
                ErrorKind::LongJump(to) => {
                    write!(f, "Jump target ({to}) outside of range -0x400..=0x3fe")
                }
                ErrorKind::Errors(errors) => {
                    writeln!(f, "Could not complete assembly:")?;
                    for error in errors {
                        writeln!(f, "{error}")?;
                    }
                    Ok(())
                }
            }
        }
    }
}
