// © 2023-2024 John Breaux
//See LICENSE.md for license
//! A bare-bones toy assembler for the TI MSP430, for use in MicroCorruption
//!
//! This project aims to assemble any valid msp430 instructions, while including important quality
//! of life features such as constant expression evaluation.
//!
//! ## Tokenization
//! The [`Lexer`](lexer::Lexer) will ignore whitespace, except newlines. It borrows a text buffer,
//! and outputs [tokens](lexer::token::Token) of various [TokenKinds](lexer::token::TokenKind).
//!
//! ## Preprocessing
//! The [`Preprocessor`](preprocessor::Preprocessor) will filter
//! [newlines](lexer::token::TokenKind::Newline), unless used to terminate a `.define` directive.
//!
//! ## Parsing
//! The [`Parser`](parser::Parser) consumes a [Lexer](lexer::Lexer)
//! and returns an [AST](parser::ast) structured roughly as follows:
//! ```text
//! Statements
//! ├─ Stmt
//! │  └─ Comment
//! ├─ Stmt
//! │  └─ Directive                 // Pre- or Post-processor directive
//! ├─ Stmt
//! │  └─ Label                     // Label definition
//! ├─ Stmt
//! │  └─ Insn
//! │     └─ NoEm                   // A zero-operand "emulated" instruction
//! ├─ Stmt
//! │  └─ Insn
//! │     └─ OneEm                  // A one-operand "emulated" instruction
//! │        ├─ Opcode
//! │        ├─ Width
//! │        └─ Dst                 // A destination register has several addressing modes:
//! │           └─ Direct           // - The contents of a register
//! │           ╶─ Indexed          // - The register, as a pointer, plus a byte index
//! │           ╶─ Absolute         // - An immediate absolute address
//! │           ╶─ Special          // - A so-called "special" immediate (#0 or #1) - these are joke encodings.
//! ├─ Stmt
//! │  └─ Insn
//! │     └─ OneArg                 // A one-operand instruction
//! │        ├─ Opcode
//! │        ├─ Width
//! │        └─ Src                 // A source register has even more addressing modes:
//! │           └─ Direct           // - The contents of a register
//! │           ╶─ Indexed          // - The register, as a pointer, plus a byte index
//! │           ╶─ Indirect         // - The word at the address stored in the register
//! │                               //   (like Indexed, but without an extension word.)
//! │           ╶─ PostIncrement    // - Indirect, but the register is post-incremented by 1
//! │                               //   (or, if it's the PC or SP, by 2)
//! │           ╶─ Absolute         // - An immediate absolute address
//! │           ╶─ Immediate        // - An immediate 16-bit number
//! │           ╶─ Special          // - A so-called "special" immediate (#0 or #1) - these are joke encodings.
//! ├─ Stmt
//! │  └─ Insn
//! │     └─ TwoArg                 // A two-operand instruction
//! │        ├─ Opcode
//! │        ├─ Width
//! │        ├─ Src
//! │        └─ Dst
//! └─ Stmt
//!    └─ Insn
//!       └─ Jump                   // A relative jump instruction
//!          ├─ Opcode              // The jump condition
//!          └─ JumpDst             // A jump instruction's destination can be either:
//!             └─ Rel              // - An even, signed 11-bit offset
//!             ╶─ Label            // - A label to jump to
//! ```
//!
//! ## Canonicalization
//! After parsing, tokens must be [canonicalized](parser::ast::canonical::Canonicalize):
//! - Expressions which act exclusively on numbers are eagerly evaluated
//!   - Expressions which begin with a numeric part are repacked for late evaluation
//! - "Emulated" instructions are desugared into their canonical counterparts
//!
//! ## Assembly
//! The [Assembler](assembler::Assembler) takes an [AST](parser::ast), and
//! 1. Encodes all [Instructions](parser::ast::Instruction) into 16-bit words
//! 2. Records all jump labels, for backpatching
//! 3. Records all expressions, for late evaluation
//! 4. Performs late evaluation and backpatching
//!
//! If a non-canonical instruction is found, the assembler will print a warning,
//! and canonicalize it.

pub mod span;

pub mod lexer;

pub mod preprocessor;

pub mod parser;

pub mod assembler;
